@@ -216,13 +216,14 @@ typedef struct _xf86CrtcFuncs {
 
     /**
      * Callback for panning. Doesn't change the mode.
+     * Added in ABI version 2
      */
     void
     (*set_origin)(xf86CrtcPtr crtc, int x, int y);
 
 } xf86CrtcFuncsRec, *xf86CrtcFuncsPtr;
 
-#define XF86_CRTC_VERSION 2
+#define XF86_CRTC_VERSION 3
 
 struct _xf86Crtc {
     /**
@@ -236,9 +237,9 @@ struct _xf86Crtc {
     ScrnInfoPtr	    scrn;
     
     /**
-     * Active state of this CRTC
+     * Desired state of this CRTC
      *
-     * Set when this CRTC is driving one or more outputs 
+     * Set when this CRTC should be driving one or more outputs 
      */
     Bool	    enabled;
     
@@ -312,18 +313,19 @@ struct _xf86Crtc {
      * Current transformation matrix
      */
     PictTransform   crtc_to_framebuffer;
-    struct pict_f_transform f_crtc_to_framebuffer;
-    struct pict_f_transform f_framebuffer_to_crtc;
-    PictFilterPtr   filter;
-    xFixed	    *params;
-    int		    nparams;
-    int		    filter_width;
-    int		    filter_height;
+    /* framebuffer_to_crtc was removed in ABI 2 */
+    struct pict_f_transform f_crtc_to_framebuffer; /* ABI 2 */
+    struct pict_f_transform f_framebuffer_to_crtc; /* ABI 2 */
+    PictFilterPtr   filter; /* ABI 2 */
+    xFixed	    *params; /* ABI 2 */
+    int		    nparams; /* ABI 2 */
+    int		    filter_width; /* ABI 2 */
+    int		    filter_height; /* ABI 2 */
     Bool	    transform_in_use;
-    RRTransformRec  transform;
-    Bool	    transformPresent;
-    RRTransformRec  desiredTransform;
-    Bool	    desiredTransformPresent;
+    RRTransformRec  transform; /* ABI 2 */
+    Bool	    transformPresent; /* ABI 2 */
+    RRTransformRec  desiredTransform; /* ABI 2 */
+    Bool	    desiredTransformPresent; /* ABI 2 */
     /**
      * Bounding box in screen space
      */
@@ -333,10 +335,28 @@ struct _xf86Crtc {
      * TotalArea: total panning area, larger than CRTC's size
      * TrackingArea: Area of the pointer for which the CRTC is panned
      * border: Borders of the displayed CRTC area which induces panning if the pointer reaches them
+     * Added in ABI version 2
      */
     BoxRec          panningTotalArea;
     BoxRec          panningTrackingArea;
     INT16           panningBorder[4];
+
+    /**
+     * Current gamma, especially useful after initial config.
+     * Added in ABI version 3
+     */
+    CARD16 *gamma_red;
+    CARD16 *gamma_green;
+    CARD16 *gamma_blue;
+    int gamma_size;
+
+    /**
+     * Actual state of this CRTC
+     *
+     * Set to TRUE after modesetting, set to FALSE if no outputs are connected
+     * Added in ABI version 3
+     */
+    Bool	    active;
 };
 
 typedef struct _xf86OutputFuncs {
@@ -574,7 +594,10 @@ struct _xf86Output {
 #else
     void		*randr_output;
 #endif
-    /** Desired initial panning */
+    /** 
+     * Desired initial panning
+     * Added in ABI version 2
+     */
     BoxRec          initialTotalArea;
     BoxRec          initialTrackingArea;
     INT16           initialBorder[4];
@@ -711,7 +734,7 @@ xf86CrtcRotate (xf86CrtcPtr crtc);
  * Clean up any rotation data, used when a crtc is turned off
  * as well as when rotation is disabled.
  */
-void
+extern _X_EXPORT void
 xf86RotateDestroy (xf86CrtcPtr crtc);
 
 /*
@@ -802,6 +825,7 @@ xf86OutputGetEDID (xf86OutputPtr output, I2CBusPtr pDDCBus);
  * Initialize dga for this screen
  */
 
+#ifdef XFreeXDGA
 extern _X_EXPORT Bool
 xf86DiDGAInit (ScreenPtr pScreen, unsigned long dga_address);
 
@@ -811,6 +835,7 @@ xf86DiDGAInit (ScreenPtr pScreen, unsigned long dga_address);
 
 extern _X_EXPORT Bool
 xf86DiDGAReInit (ScreenPtr pScreen);
+#endif
 
 /*
  * Set the subpixel order reported for the screen using
@@ -900,18 +925,10 @@ extern _X_EXPORT void
 xf86_crtc_notify(ScreenPtr pScreen);
 
 /**
- * Panning
+ * Gamma
  */
-Bool
-xf86_crtc_get_panning(ScrnInfoPtr pScrn,
-		      BoxPtr      totalArea,
-		      BoxPtr      TrackingArea,
-		      INT16      *border);
 
-Bool
-xf86_crtc_set_panning(ScrnInfoPtr pScrn,
-		      BoxPtr      totalArea,
-		      BoxPtr      TrackingArea,
-		      INT16      *border);
+extern _X_EXPORT Bool
+xf86_crtc_supports_gamma(ScrnInfoPtr pScrn);
 
 #endif /* _XF86CRTC_H_ */
